#include "CompBioTools.hh"
#include "AceFile.hh"

// An HGSC specific class for dealing with scaffolds. 
// Class can be found under HGSC++. 
#include "Scaffold.hh"

string gVersion = "1.40";

int GetContigNum(AceFile &Ace,string cName)
{
  for(int c = 0;c< Ace.size();c++){
	 if (cName == Ace[c].mName){
		return(c);
	 }
  }
  return(-1);
}

//---------------------------------------------------
// * PrintVersion
//---------------------------------------------------
//
void PrintVersion()
{
  cout<<"
    Version "<<gVersion<<endl;
}

string OptScaffoldFileName = "";
string OptAceFileName = "";
string OptOutFileName = "";


int OptMinScaffoldSize = 15000;
int OptMinContigSize = 1000;
int OptMinLongLink = 50000;

int OptMaxGapSize = 0;
int OptMaxSeqSize = 0;

int OptQualityTrim = 1;
int OptMinN = 50;
int OptNumX = 100;
bool OptDebug = false;
bool OptFixedN = false;
bool OptMultiFileMode = false;
bool OptSingleSequence = false;
bool OptQualOut = false;
bool OptIncludeCriteriaUnscaffolded = false;
string OptProjectName ="";

int OptBigTemplateSize = 50000;


//---------------------------------------------------
// * OneNeighborAndOnlyLongLinks
//---------------------------------------------------
//
bool OneNeighborAndOnlyLongLinks(AceFile &Ace,string ContigName,ScaffoldGraph &G){

  // Find the neighbors of contig c in the scaffold. 
  string name1="",name2="";
  G.FindNeighbors(ContigName,name1,name2);
 
  int c = GetContigNum(Ace,ContigName);
  if (c<0){
	 cerr<<ContigName<<" does not appear to be in the ace file."<<endl;
  }
 
  //	 cout<<Ace[c].mName<<"-->\t"<<name1<<"\t"<<name2<<endl;
  
  int neighbor1 =-1,neighbor2 = -1; 
  if (name1 != "") neighbor1 = Ace.CName2CIdx(name1);
  if (name2 != "") neighbor2 = Ace.CName2CIdx(name2);

  // Then see if the templates shared between these contigs are all long inserts or not. 
  // First get the template info for templates that span the given contigs. 
  vector<TemplateInfo> Templates1,Templates2;
  

  // KJD BUG:  This falsely assumes that the links will be to the adjacent 
  // KJD BUG:  contig!  A transitive link might connect it, however.  Need 
  // KJD BUG:  to do some more global comparison. 
  Ace.FindTemplatesBetween(c,neighbor1,Templates1);
  Ace.FindTemplatesBetween(c,neighbor2,Templates2);

  cerr<<Ace[c].mName<<"\t";
  
  bool bAllLongLinks = true;
  for(int i = 0;i< Templates1.size();i++){
	 //cout<<"\tTemplates[i].mLength="<<Templates[i].mLength<<endl;
	 if (Templates1[i].mInsertSize < OptBigTemplateSize){
		bAllLongLinks = false;
	 }
  }

  cout<<"Templates1 <OptBig:"<<bAllLongLinks<<"\t";

  for(int i = 0;i< Templates2.size();i++){
	 //cout<<"\tTemplates[i].mLength="<<Templates[i].mLength<<endl;
	 if (Templates2[i].mInsertSize < OptBigTemplateSize){
		bAllLongLinks = false;
	 }
  }
  cout<<"Templates2<OptBig:"<<bAllLongLinks<<"\t";
  
  // If there are no templates, then we want to return true...
  // If there are no templates then this doesn't have "only long links". 
  if ((Templates2.size() == 0) && (Templates1.size()==0)){
	 //bAllLongLinks = false;
	 // It is suspicious, however
	 cerr<<"Why doesn't scaffolded contig: "<<Ace[c].mName<<" have any links?"<<endl;
  }
  //cout<<"No templates:"<<bAllLongLinks<<"\t";

  // If there are two neighbors, it's not "one neighbor".   
  if ((Templates1.size() >0) && (Templates2.size() > 0)) bAllLongLinks = false;

  cout<<name1<<"-"<<name2<<" ";
  cout<<"("<<Templates1.size()<<","<<Templates2.size()<<") ";
  cout<<"TwoNeighbors:"<<bAllLongLinks<<endl;

  return(bAllLongLinks);
}

//---------------------------------------------------
// * PrintHelp
//---------------------------------------------------
//
void PrintHelp()
{
printf("\n
    atlas-refsequence -a AceFile -g ScaffoldGraph -o Output {Options}

    Reads in an ace file and an associated graph file and produces 
    a reference sequence.  Sequence is written out in the relative 
    order and orientation given by the graph file.  Gaps between 
    ordered and oriented sequences are padded with Ns.  In SingleSequence 
    mode Gaps between unordered sequences are padded with Xs:

    > Scaffold0
    sssssNNNNNsssssNNNNssssXXXXXsssssNNNNsssssXXXssssXXXssssXXX


    In multiple-sequence mode, each scaffold is one seqquence like:

    >Scaffold0
    sssssNNNNNsssssNNNNssss
    >Scaffold1
    sssssNNNNsssss
    >UnscaffoldedContig2
    ssss
    >UnscaffoldedContig3
    ssss

    Output is to a file that starts with ERROR if the sequence ends
    up with blocks of > 50K Ns. 


    -a   Ace file name. 
    -g   Scaffold .graph file name. 
    -o   Output file name. 

    -p   Project name to append to _Scaffold1, etc. 
    ----

    -q   Also write quality file out to Output.qual

    -S   Enable single sequence mode. Default is multiple sequence.

    -m   Multiple *file* mode.   Each scaffold written to it's own file. 

    -F   Use fixed number of Ns to pad between scaffolds rather than 
         based on distance estimate.  

    -s   (15000) Minimum size of scaffold to output. 

    
    -U   Include unscaffolded contigs (that meet criteria for unscaffolded contigs)
    -M   Minimum size for BAC only unscaffolded contigs. 

    -Q   (1)    Quality trim for contigs.  

    -X   (100)  Number of X's to pad between unordered sequences. 
    -N   (50)   Minimum number of Ns to pad between ordered seqs. 
 
    -G	MaxGapSize to allow in output. Reports an error and writes output to 
	      .FAIL file if gap greater than this in output.  Default is disabled. 

    -O  MaxOutputSize.  Reports an error and writes output to .FAIL file if 
        the total sequence size exceeds this ammount.  Default is disabled.

    -D   Show debugging output (Contig and gap structure printed). 

    Written by: James Durbin (kdurbin@bcm.tmc.edu)
");

 PrintVersion();

 printf(" 
    Version History:  
    ----------------
    1.40   Fixed bug where GetTemplateInfo would just add templates on multiple 
           calls, resulting in an explosion in RAM and run time. 
    1.32   Fixed serious bug in code that finds shared contigs between templates. 
           This bug made most pairs seem template poor, resulting in some dropped contigs. 
    1.30   Fixed but in end trimming.  Added sanity checking.  Added status output.  
    1.28   Fixed bug that prevented trimming of contigs with no links that 
           appear in the scaffold.  Question is, how can it be in the 
           scaffold without some links??
    1.26   Added support for more complex evaluation. 
    1.24   Added support for writing quality out also. 
    1.22   Made coordinates 1 based.  Fixed scaffold size. 
    1.20   
    1.00   Initial release. 
");
}

//---------------------------------------------------
// * SetupOptions
//---------------------------------------------------
//
void SetupOptions(int argc, char**argv)
{
  // Print options for posterity
  for(int i = 0;i< argc;i++){
	 cout<<argv[i]<<" ";
  }
  cout<<endl;


  // Create parser object and define valid options
  // Options with a : require a parameter. 
  GetOption Parser(argc,argv,"Uqs:mSa:g:o:M:Q:N:X:p:FDhG:O:");

  // Handle the options...
  char theOption;
  while((theOption = Parser.NextOption()) != OPTION_EOF){
    switch(theOption){
		case 'a': OptAceFileName = Parser.GetStringArg();break;
		case 'g': OptScaffoldFileName = Parser.GetStringArg();break;
		case 'p': OptProjectName = Parser.GetStringArg();break;
		case 'o': OptOutFileName = Parser.GetStringArg();break;
			
		case 's': OptMinScaffoldSize = Parser.GetIntArg();break;
		case 'M': OptMinContigSize = Parser.GetIntArg();break;
		case 'Q': OptQualityTrim = Parser.GetIntArg();break;	 
			
		case 'X': OptNumX = Parser.GetIntArg();break;
		case 'N': OptMinN = Parser.GetIntArg();break;
		case 'D': OptDebug = true;break;
		case 'q': OptQualOut = true;break;
		case 'm': OptMultiFileMode = true;break;
		case 'U': OptIncludeCriteriaUnscaffolded = true;break;
		case 'S': OptSingleSequence = true;break;
		case 'F': OptFixedN = true;break;
			
		case 'G': OptMaxGapSize = Parser.GetIntArg();break;
		case 'O': OptMaxSeqSize = Parser.GetIntArg();break;

		default: PrintHelp();exit(1);break;
    }
  }
}

//-------------------------------------------
// * GetContigName
//-------------------------------------------
//
string GetContigName(string nodeName)
{
  int usIdx = nodeName.find("_");
  string reval = nodeName.substr(0,usIdx);
  return(reval);
}

//-------------------------------------------
// * IsRight
//-------------------------------------------
//
bool IsRight(string nodeName)
{
  if (Contains(nodeName,"right")) return(true);
  else return(false);
}

//-------------------------------------------
// * AddN
//-------------------------------------------
//
void AddN(Sequence &Seq,int NumN){ 
  for(int i = 0;i< NumN;i++){
	 Seq.push_back('N');
  }
}

//-------------------------------------------
// * AddZeros
//-------------------------------------------
//
void AddZeros(SequenceQuality &SeqQual,int NumN){ 
  for(int i = 0;i< NumN;i++){
	 SeqQual.push_back(0);
  }
}

//-------------------------------------------
// * AddX
//-------------------------------------------
//
void AddX(Sequence &Seq,int NumX){
  for(int i = 0;i< 100;i++){
	 Seq.push_back('X');
  }
}

//-------------------------------------------
// * GetContigSequence
//-------------------------------------------
// Gets the trimmed and depadded sequence for the named
// contig. 
// 
void GetContigSequence(AceFile &Ace,string CName,Sequence &Seq,SequenceQuality &Qual)
{
  for(int i = 0;i<Ace.size();i++){
	 if (Ace[i].mName == CName){
		Seq.mName = Ace[i].mName;
		for(int j = Ace[i].mTrimLeft;j <=Ace[i].mTrimRight;j++){
		  if (Ace[i][j] != '*'){
			 Seq.push_back(Ace[i][j]);
			 Qual.push_back(Ace[i].mConsensusQuality[j]);
		  }
		}
		return;
	 }
  }

  cout<<"ERROR:  Didn't find "<<CName<<endl;
  for(int i = 0;i< Ace.size();i++){
	 cout<<"\t"<<Ace[i].mName<<endl;
  }
}

bool OnlyLongLinks(vector<TemplateInfo> &T,int MinLongLink){
  for(int i = 0;i< T.size();i++){
	 // Go by insert size unless we have a recorded length. 
	 if (T[i].mLength <= 0){
		if (T[i].mInsertSize < MinLongLink) return(false);
	 }else{
		if (T[i].mLength < MinLongLink) return(false);
	 }
  }

  // If no small links found, must be only long links...
  return(true);
}

//-------------------------------------------
// * ScaffoldMeetsCriteria
//-------------------------------------------
//
bool ScaffoldMeetsCriteria(Scaffold &S)
{
  // Skip scaffolds that are single contigs. 
  if (S.size() == 1) {
	 if (OptDebug) cout<<"Scaffold "<<S.mID<<" NotCriteria: size == 1. "<<endl;
	 return(false);
  }

  // Skip scaffolds that are too small.
  int ScaffoldSize = (int)(S[S.size()-1].mEnd - S[0].mStart);  
  if (ScaffoldSize < OptMinScaffoldSize) {
	 if (OptDebug) cout<<S.mID<<" ScaffoldSize="<<ScaffoldSize<<" < "<<OptMinScaffoldSize<<endl;
	 return(false);
  }

  return(true);
}


//-------------------------------------------
// * ContigMeetsCriteria
//-------------------------------------------
//
bool ContigMeetsCriteria(bool bOneNeighbor, AceFile &A,string ContigName)
{
  int cIdx = A.CName2CIdx(ContigName);

	if (cIdx < 0){
		cerr<<"ERROR: No contig: "<<ContigName<<" found. "<<endl;
		return(false);
	}

  Contig &C = A[cIdx];

  vector<int> ConnectedContigs;
  A.GetContigsSharingTemplates(cIdx,ConnectedContigs);
	A[cIdx].mTInfo.clear(); // KJD See if this is where memory leak is. 

	if (OptDebug){
		cout<<endl;
		cout<<A[cIdx].mName<<" shares templates with:"<<endl;
		for(int i = 0;i< ConnectedContigs.size();i++){
			cout<<"\t"<<A[ConnectedContigs[i]].mName<<endl;
		}
	}
  
  bool bOnlyLongLinks = true;
  int numLinks = 0;
  for(int c = 0;c< ConnectedContigs.size();c++){
	 vector<TemplateInfo> Templates;
	 A.FindTemplatesBetween(cIdx,ConnectedContigs[c],Templates);

	 if (OptDebug){
		 cout<<endl;
		 cout<<"Tempates between "<<A[cIdx].mName<<" and "<<A[c].mName<<endl<<endl;
		 for(int i = 0;i< Templates.size();i++){
			 Templates[i].Print(cout);
		 }
	 }


	 numLinks+=Templates.size();
	 if (!OnlyLongLinks(Templates,OptMinLongLink)){
		bOnlyLongLinks = false;
	 }
  } 

  //  bool bWGSOnly;
  //if (C.AllReadsMatch("gnl|ti|")) bWGSOnly = true;
  //else bWGSOnly = false;
  
  if (bOneNeighbor && bOnlyLongLinks){		
		if (OptDebug) cout<<ContigName<<" FAIL  ";
		if (OptDebug) cout<<"OneNeighbor="<<bOneNeighbor<<" OnlyLong="<<bOnlyLongLinks<<endl;
		if (OptDebug) cout<<"\tMin Long="<<OptMinLongLink<<"\tnumLinks="<<numLinks<<endl;
		if (OptDebug) cout<<"\tConnectedContigs.size="<<ConnectedContigs.size()<<endl;

		return(false);
	}else {
		if (OptDebug) cout<<ContigName<<" PASS "<<endl;
		return(true);
	}

}

//-------------------------------------------
// * RemoveNonCriteriaContigs
//-------------------------------------------
// Rewrite scaffold to only include criteria contigs. 
//
void RemoveNonCriteriaContigs(ScaffoldGraph &G,AceFile &A)
{
	ScaffoldGraph GOut;

	//	cout<<"Graph size="<<G.size()<<endl;
	for(int s = 0;s< G.size();s++){
		Scaffold &S = G[s];		
		Scaffold SOut;
		SOut.mID = S.mID;
		
		// Only add the criteria contigs...
		for(int c = 0;c< S.size();c++){
			bool bOneNeighbor = false;
			if ((c == 0) || (c == S.size()-1)) bOneNeighbor = true;
			cout<<"Check if "<<S[c].mName<<" meets criteria..."<<flush;
			if (ContigMeetsCriteria(bOneNeighbor,A,S[c].mName)){
				cout<<"YES.  Done."<<endl;
				SOut.push_back(S[c]);
			}else{
				cout<<"NO.  Done."<<endl;
			}

		}


		
		// Recompute the size...
		if (SOut.size() > 1){
			SOut.mSize = SOut[SOut.size()-1].mEnd - SOut[0].mStart +1;
		}else{
			if (SOut.size() == 1){
				SOut.mSize = SOut[0].mEnd - SOut[0].mStart + 1;
			}else{
				SOut.mSize = 0;
			}
		}	
		// Add it to the set of output scaffolds...
		GOut.push_back(SOut);
	}

	// Replace input graph with the newly created graph. 
	G = GOut;

}


//-------------------------------------------
// * GetScaffoldSeq
//-------------------------------------------
// Adds sequence for the contigs in this scaffold (if good) to output. 
// 
void GetScaffoldSeq(Scaffold &S,AceFile &A,
						  Sequence &ScaffoldSeq,
						  SequenceQuality &ScaffoldQual,
						  Scaffold &SActual)
{
  int seqNum = 0;
  bool bFirstTime = true;
  
  ScaffoldSeq.mName = S.mID;
  ScaffoldQual.mName = S.mID;
  if (ScaffoldMeetsCriteria(S)){
	 
	 if (OptDebug) cout<<"\nScaffold "<<S.mID<<endl;
	 
	 SActual.mID = S.mID;
	 SActual.mSize = S.mSize;
	 
	 // vector<Sequence> CriteriaContigs;
	 //vector<SequenceQuality> CriteriaQual;
	 // Record each criteria contig in the scaffold...
	 int aStart = 0;
	 for(int c = 0;c < S.size();c++){
		
		 //		bool bOneNeighbor = false;
		 //if ((c == 0) || (c == S.size()-1)){
		 //bOneNeighbor=true;
		 //}

		 //if (ContigMeetsCriteria(bOneNeighbor,A,S[c].mName)){
		 //		 cout<<S[c].mName<<" meets criteria."<<endl;
		 
		 // Get the seq and qual from the Ace file (trimmed and depadded). 
		 Sequence Contig;
		 SequenceQuality Qual;
		 GetContigSequence(A,S[c].mName,Contig,Qual);
		 
		 // Orient the sequence
		 if (S[c].mSense=='-') {
			 Contig.ReverseComplement();
			 Qual.Reverse();
		 }
		 
		 // Compute the true contig boundaries for a true scaffold graph. 
		 ScaffoldContig CTemp;
		 CTemp.mStart = CTemp.mEnd = 0;
		 CTemp.mName = S[c].mName;
		 CTemp.mSense = S[c].mSense;
		 CTemp.mStart = aStart;
		 CTemp.mEnd = aStart+Contig.size()-1;
		 SActual.push_back(CTemp);
		 
		 aStart = (int) CTemp.mEnd+1;
		 
		 // Add this sequence to the scaffold sequence. 
		 
		 if (OptDebug) cout<<"Contig: "<<Contig.mName<<"\t"<<Contig.size()<<"\t"<<S[c].mSense<<endl;
		 
		 ScaffoldSeq.Concatenate(Contig);
		 ScaffoldQual.Concatenate(Qual);
		 
		 // Add N's when it's not the last contig in the scaffold...
		 if (c < S.size()-1){
			 
			 // Only add the Ns if the next contig is not omitted...
			 // if (((c+1) == 0) || ((c+1) == S.size()-1)){
			 // bOneNeighbor=true;
			 // }
			 //if (ContigMeetsCriteria(bOneNeighbor,A,S[c+1].mName)){
			 int gapSize = (int)( S[c+1].mStart - S[c].mEnd);
			 int NumN = OptMinN;
			 if (!OptFixedN) NumN = max(gapSize,OptMinN);
			 if (OptDebug) cout<<"NNN * "<<NumN<<endl;
			 
			 aStart=aStart+NumN;
			 
			 AddN(ScaffoldSeq,NumN);
			 AddZeros(ScaffoldQual,NumN);
		 }
	
		 if (OptDebug) cout<<"EndOfScaffold"<<endl<<endl;
	 }
	}else{
		if (OptDebug) cout<<"Scaffold "<<S.mID<<"\tdoes not meet criteria."<<endl;
	}
}
	 

//-------------------------------------------
// * AddContig
//-------------------------------------------
// Add unscaffolded contig to the output sequence. 
// NOTE: Ideally this would look at T7 SP6 markers and 
// stuff this unscaffolded sequence between those, but 
// for now it just tacks it onto the end. 
// 
void AddContig(string CName,AceFile &A, SequenceVector &SeqsOut,SequenceQualityVector &QualOut)
{
  int seqNum = SeqsOut.size() -1;
  if (OptSingleSequence){
	 AddX(SeqsOut[seqNum],OptNumX);
	 AddZeros(QualOut[seqNum],OptNumX);
  }else{	 
	 SeqsOut.resize(SeqsOut.size()+1);
	 QualOut.resize(QualOut.size()+1);
	 seqNum++;
	 SeqsOut[seqNum].mName = "UnscaffoldedContig"+itoa(seqNum);
	 QualOut[seqNum].mName = "UnscaffoldedContig"+itoa(seqNum);
  }

  // Figure out which contig CName comes from
  int c = GetContigNum(A,CName);
  
  bool bWGSOnly = A[c].AllReadsMatch("gnl|ti|");
  bool bBACOnly = A[c].NoReadsMatch("gnl|ti|");
  
  Sequence ContigSeq;
  SequenceQuality ContigQual;
  GetContigSequence(A,CName,ContigSeq,ContigQual);
  
  // Only save unscaffolded contigs if they are not WGSOnly and greater 
  // than the minimum contig size. 
  if ((ContigSeq.size() >= OptMinContigSize) && (!bWGSOnly)){
	 for(int i = 0; i< ContigSeq.size();i++){
		SeqsOut[seqNum].push_back(ContigSeq[i]);
		QualOut[seqNum].push_back(ContigQual[i]);
	 }
	 if (OptDebug) cout<<CName<<"\t"<<ContigSeq.size()<<endl;
  }
}

//-------------------------------------------
// * GetUnscaffoldedNames
//-------------------------------------------
// Figure out which contigs are not in any scaffold. 
// 
void GetUnscaffoldedNames(vector<string> &CNames,ScaffoldGraph &G,
								  AceFile &Ace)
{
  // Record the scaffolded contigs. 
  StringSet ScaffoldedContigs;
  for(int s = 0;s< G.size();s++){
	 Scaffold &S = G[s];
	 // Scaffolds which contain only one contig are "unscaffolded" 
	 if (S.size() > 1){
		for(int c = 0;c< S.size();c++){
		  ScaffoldContig &C = S[c];
		  ScaffoldedContigs.Add(C.mName);
		}
	 }
  }
  
  // Find all the contigs in the ace file which are not
  // listed as being scaffolded. 
  for(int i = 0;i< Ace.size();i++){
	 if (!ScaffoldedContigs.Contains(Ace[i].mName)){		
		CNames.push_back(Ace[i].mName);
	 }
  }
}

//-------------------------------------------
// * FixUpNames
//-------------------------------------------
// Transforms names into format like:
//
// atlas_gaaa_scaffold1 20020819 95031bp CH230-2J24 Contig61_Contig53
//
void FixUpNames(AceFile &A, string AceName, string ProjectName,
					 SequenceVector &SeqsOut2,
					 SequenceQualityVector &QualOut2){
  
  string dateStr = GetDateStamp(AceName);
  
  if (SeqsOut2.size() != QualOut2.size()){
	 cerr<<"ERROR: SeqsOut2.size="<<SeqsOut2.size()<<"\tQualOut2.size()="<<QualOut2.size()<<endl;
	 exit(1);
  }  
  
  string cloneName = "CLONENAME";

  for(int i = 0;i< SeqsOut2.size();i++){
	 string OldName = SeqsOut2[i].mName;
	 string NewDesc = dateStr;
	 NewDesc+="\t";
	 string sizeStr = itoa(SeqsOut2[i].size());
	 NewDesc+=sizeStr;
	 NewDesc+="bp\t";
	 NewDesc+=cloneName+"\t"+OldName;
	 
	 SeqsOut2[i].mDescription = NewDesc;
	 QualOut2[i].mDescription = NewDesc;
	 
	 string NewName = "atlas_"+OptProjectName+"_Scaffold"+itoa(i);
	 SeqsOut2[i].mName = NewName;
	 QualOut2[i].mName = NewName;
  }
}

#define EXIT_OK 0
#define TOO_LARGE 1
#define BIG_GAP   2
#define TOO_LARGE_AND_BIG_GAP 3
#define SEQ_QUAL_SIZEMISMATCH 4
#define INPUT_FILE_NOT_FOUND  5

//-------------------------------------------
// * PassSanityTest
//-------------------------------------------
//
// Make sure seq and quality sizes match and 
// make sure there are no long runs of N's. 
//
bool PassSanityTest(SequenceVector &SeqVec,SequenceQualityVector &QualVec,int &errCode)
{
	// Disable sanity testing. 
	if (OptMaxSeqSize == 0) return(true);
	if (OptMaxGapSize == 0) return(true);

  int MaxNBlock = OptMaxGapSize;
  int TotalSize = 0;

  if (SeqVec.size() != QualVec.size()){
		errCode = SEQ_QUAL_SIZEMISMATCH;
		return(false);
	}

	bool bFoundBigGap = false;
  for(int i = 0;i< SeqVec.size();i++){
	 if (SeqVec[i].size() != QualVec[i].size()){
		 errCode = SEQ_QUAL_SIZEMISMATCH;
		 return(false);
	 }

	 Sequence &S = SeqVec[i];
	 
	 TotalSize += S.size();

	 bool bInRun = false;
	 int NCount = 0;
	 for(int j = 0;j < S.size();j++){		

		if (toupper(S[j]) == 'N'){
		  if (bInRun) NCount++;
		  else{
			 NCount = 0;
			 bInRun = true;
		  }
		}else{
		  // If this is the end of a run
		  if (bInRun){
			 if (NCount > MaxNBlock) {
				 cerr<<"ERROR NCount = "<<NCount<<endl;
				 bFoundBigGap = true;
			 }
			 bInRun = false;
			 NCount = 0;
		  }		  
		}
	 }
	 // We might reach the end of the sequence without 
	 // hitting a non-N. 
	 if (NCount > MaxNBlock){
		 cerr<<"ERROR NCount = "<<NCount<<endl;
		 bFoundBigGap = true;
		 return(false); 
	 }
  }

	if (bFoundBigGap){
		errCode = BIG_GAP;		
		if (TotalSize > OptMaxSeqSize) errCode = TOO_LARGE_AND_BIG_GAP;
		return(false);
	}


	if (TotalSize > OptMaxSeqSize){
		errCode = TOO_LARGE;
		return(false);
	}

  return(true);
}


//-------------------------------------------
// * SplitScaffoldsOnBigGap
//-------------------------------------------
//
bool SplitScaffoldsOnBigGap(ScaffoldGraph &GraphIn)
{

	bool bSplit = false; // assume no splits until find one. 

	// Disable this when no max gap size given. 
	if (OptMaxGapSize == 0) return(false);

	ScaffoldGraph GraphOut;

	for(int s = 0;s < GraphIn.size();s++){
		Scaffold &S = GraphIn[s];
		Scaffold SOut; 
	
		S.SortByStart();
		if (S.size() > 1){
			int lastc = 0;	
			SOut.mID = S[0].mName; // First part of ID. 
			for(int c = 0;c < S.size() - 1;c++){

				int s2 = (int) S[c+1].mStart;
				int e1 = (int) S[c].mEnd;
				
				int gap = s2-e1;
				if (gap > OptMaxGapSize){
					bSplit = true;

					// So, this finishes off that scaffold, add it to the graph. 
					SOut.push_back(S[c]);
					SOut.mID=SOut.mID+"_"+S[c].mName;
					SOut.ComputeSize();
					GraphOut.push_back(SOut);

					// Start a new scaffold for the remaining contigs. 
					SOut.clear();
					SOut.mID = S[c+1].mName;
				}else{
					// Keep adding contigs to the scaffold. 
					SOut.push_back(S[c]);
				}
			}
			
			if (SOut.size() > 0){
				if (SOut.size() > 1) SOut.mID=SOut.mID+"_"+SOut[SOut.size()-1].mName;
				SOut.ComputeSize();
				GraphOut.push_back(SOut);
				SOut.clear();
			}
		}else{
			GraphOut.push_back(S);
		}
	}
	GraphIn = GraphOut;

	return(bSplit);

}

//-------------------------------------------
// * AddAllScaffoldSequences
//-------------------------------------------
//
void AddAllScaffoldSequences(ScaffoldGraph &G,AceFile &Ace, SequenceVector &SeqsOut,
														 SequenceQualityVector &QualOut,ScaffoldGraph &GoodGraph)
{

  if (OptSingleSequence){
		SeqsOut.resize(1);
		QualOut.resize(1);
  }
	
	cout<<"Graph.size="<<G.size()<<endl<<endl;
  for(int s = 0;s< G.size();s++){	
		Sequence ScaffoldSeq;
		SequenceQuality ScaffoldQual;
		Scaffold GoodScaffold;
		
		GetScaffoldSeq(G[s],Ace,ScaffoldSeq,ScaffoldQual,GoodScaffold);
		
		if (OptSingleSequence){
			SeqsOut[0].Concatenate(ScaffoldSeq);
			QualOut[0].Concatenate(ScaffoldQual);
			
			if (OptDebug) cout<<"\n"<<ScaffoldSeq.mName<<"\t"<<ScaffoldSeq.size()<<endl;
			if (s < G.size() -1){		  
				if (OptDebug) cout<<"XXX "<<OptNumX<<endl;
				AddX(SeqsOut[0],OptNumX);
				AddZeros(QualOut[0],OptNumX);
			}
		}else{
			if (OptDebug) cout<<ScaffoldSeq.mName<<"\t"<<ScaffoldSeq.size()<<endl;
			SeqsOut.push_back(ScaffoldSeq);
			QualOut.push_back(ScaffoldQual);
		}
		GoodGraph.push_back(GoodScaffold);
  }
}
	
//-------------------------------------------
// * AddAllUnScaffoldContigs
//-------------------------------------------
//
void AddAllUnscaffoldedContigs(ScaffoldGraph &G,
															 AceFile &Ace,
															 SequenceVector &SeqsOut,
															 SequenceQualityVector &QualOut)
{
	if (OptDebug) cout<<"--------"<<endl;
  vector<string> CNames;

  // Add unscaffolded contigs if they meet the criteria
  if (OptIncludeCriteriaUnscaffolded){
		GetUnscaffoldedNames(CNames,G,Ace);
		for(int i = 0;i< CNames.size();i++){
			AddContig(CNames[i],Ace,SeqsOut,QualOut);
		}
  } 
}


//-------------------------------------------
// * ComputeSize
//-------------------------------------------
//
void ComputeSize(int &BaseSize,int &totalSize,SequenceVector &SeqsOut)
{
	BaseSize = 0;
	totalSize = 0;
	for(int j = 0;j < SeqsOut.size();j++){
		totalSize+=SeqsOut[j].size();
		for(int i = 0;i< SeqsOut[j].size();i++){
			if ((SeqsOut[j][i] != 'N') && (SeqsOut[j][i] != 'X')) BaseSize++;
		}
  }
}

//-------------------------------------------
// * RemoveZeroLengthSequences
//-------------------------------------------
//
void RemoveZeroLengthSequences(SequenceVector &SeqsOut,SequenceQualityVector &QualOut){		
	SequenceVector SeqsOut2;
	SequenceQualityVector QualOut2;
	for(int i = 0;i< SeqsOut.size();i++){
		if (SeqsOut[i].size() > 0){
			SeqsOut2.push_back(SeqsOut[i]);
		}
	}
	for(int i = 0;i< QualOut.size();i++){
		if (QualOut[i].size() > 0){
			QualOut2.push_back(QualOut[i]);
		}
	}
	SeqsOut = SeqsOut2;
	QualOut = QualOut2;
}

//-------------------------------------------
// * SaveFiles
//-------------------------------------------
//
void SaveFiles(SequenceVector &SeqsOut,SequenceQualityVector &QualOut,
							 ScaffoldGraph &GoodGraph)
{

  // Save the desired files... 
  FASTAFile FA;
  if (!OptMultiFileMode){
		cout<<"Writing "<<OptOutFileName<<"..."<<flush;
		FA.Write(OptOutFileName,SeqsOut);
		
		if (OptQualOut){
			string qName = OptOutFileName+".qual";
			FA.Write(qName,QualOut);
		}

		string qName = OptOutFileName+".scaffold";
		cout<<"GoodGraph.size="<<GoodGraph.size()<<endl;
		GoodGraph.Write(qName);

		cout<<"done."<<endl;
  }else{
		for(int s = 0;s< SeqsOut.size();s++){
			string outName = OptOutFileName+SeqsOut[s].mName+".linear.fa";
			string qName = outName+".qual";
			FA.Write(outName,SeqsOut[s]);
			if (OptQualOut) FA.Write(qName,QualOut[s]);
		}

		for(int s = 0;s< GoodGraph.size();s++){
			string outName = OptOutFileName+GoodGraph[s].mID+".linear.scaffold";
			GoodGraph[s].Write(outName);
		}
  }
}


//-------------------------------------------
// * main
//-------------------------------------------
//
int main(int argc,char **argv)
{
  int exitStatus = 0;

  SetupOptions(argc,argv);
  
  if (OptDebug) cout<<"Debugging on."<<endl;

  SequenceVector SeqsOut,BadSeqsOut;
  SequenceQualityVector QualOut,BadQualOut;

  ScaffoldGraph GoodGraph,BadGraph;
  
  ScaffoldGraph G;
  cout<<"Reading "<<OptScaffoldFileName<<"..."<<flush;
  G.Read(OptScaffoldFileName);
  cout<<G.size()<<" scaffolds read."<<endl;


	if (OptDebug){
		for(int s = 0;s< G.size();s++){
			cout<<"Scaffold "<<G[s].mID<<" size="<<G[s].size()<<endl;
		}
	}

    
  AceFile Ace;
  Ace.mContigTrim = OptQualityTrim;
  Ace.mPhredTrim = 20;
  cout<<"Reading "<<OptAceFileName<<"..."<<flush;
  Ace.Read(OptAceFileName);
  cout<<Ace.size()<<" Ace contigs read."<<endl;

	// Clean up scaffold to include only criteria contigs 
	cout<<"Removing non-criteria contigs..."<<endl;
	RemoveNonCriteriaContigs(G,Ace);
	cout<<"done."<<endl;

	// Split any scaffolds that have gaps bigger than 50Kbp. 
	//bool bSplitScaffolds = SplitScaffoldsOnBigGap(G);
	bool bSplitScaffolds = false;

	// Copy the sequence for each cleaned up scaffold, putting
	// order and orientation right...
	cout<<"Add all scaffold sequences..."<<flush;
	AddAllScaffoldSequences(G,Ace,SeqsOut,QualOut,GoodGraph);
	cout<<"done."<<endl;

  // Add the unscaffolded contigs to the output
	cout<<"Add all unscaffolded contigs..."<<flush;
	AddAllUnscaffoldedContigs(G,Ace,SeqsOut,QualOut);
	cout<<"done."<<endl;

	int BaseSize=0,totalSize=0;
	ComputeSize(BaseSize,totalSize,SeqsOut);
  cout<<"Num sequences = "<<SeqsOut.size()<<endl;
  cout<<"Total size = "<<totalSize<<endl;
  cout<<"Size without N or X = "<<BaseSize<<endl;

  // Delete any zero length sequences.  KJD Why would this happen??
	RemoveZeroLengthSequences(SeqsOut,QualOut);

  // Rename the output for external export. 
  if (OptProjectName != ""){
	 FixUpNames(Ace,OptAceFileName,OptProjectName,SeqsOut,QualOut);
  }
    
	// Check that the output meets minimum sanity criteria. Report an 
	// exit error if there is something wrong. 
  if (!PassSanityTest(SeqsOut,QualOut,exitStatus)){
  	 cerr<<"Sanity test failed.  Output written to err file."<<endl;
  	 OptOutFileName=OptOutFileName+".FAIL";
		 if ((exitStatus == TOO_LARGE) && (bSplitScaffolds)){
			 exitStatus = TOO_LARGE_AND_BIG_GAP;
		 }
  }

	// Save the file or files...
	SaveFiles(SeqsOut,QualOut,GoodGraph);

	cout<<"exitStatus = "<<exitStatus<<endl;
  exit(exitStatus);
}





