//==============================================================================
//	 
// phh-extract-reads.cc
// Inspired by KJD's initial version.
//
//==============================================================================
// $Header: //CBT-depot/CompBioToolsApps/tools/atlas-extract-reads.cc#1 $
//------------------------------------------------------------------------------
// $Log: phh-extract-reads.cc,v $
// Revision 1.2  2000/08/14 13:57:57  havlak
// Added features for controlling how many sequences go into each output
// file -- for producing multiple output files.
//
// Revision 1.1  2000/07/07 18:28:03  havlak
// Initial revision
//
//
//------------------------------------------------------------------------------
//.......1.........2.........3.........4.........5.........6.........7.........8
//

#include "LocalSTL.hh"
#include "GetOption.hh"
#include "Sequence.hh"
#include "FASTAFile.hh"
#include "FileUtils.hh"
#include <hash_set>

void myExit(int exitVal) {
  cout << flush;
  cerr << flush;
  exit(exitVal);
}

string OptFastaName;
string OptFileOfDescrip;
string OptOutFile;
string OptDebug = "";
long OptSeqPerFile = 0;
bool OptMultiSeq = false;
bool OptExclude = false;
string OptSeparators = " ";
static void closeOut();
static void openNext();

bool debugging(const char which[]) {
  if (OptDebug.find('+') != string::npos) return true;

  for (int i = 0; which[i]; i++) {
    if (OptDebug.find(which[i]) != string::npos) return true;
  }
  return false;
}

//---------------------------------------------------
// * PrintOptions
//---------------------------------------------------
//
void PrintOptions(string prefix)
{
  cout << prefix << " options are in []:\n"<<
    "   -i {FASTAFile}        ["<< OptFastaName <<"] Name of FASTA (sequence or quality) input file.\n"<<
    "   -f {FileOfDescrips}   ["<< OptFileOfDescrip << "] File read descriptions to extract.\n"<<
    "   -o {OutFile}          ["<< OptOutFile << "] Name of FASTA output file.\n"<<
    "   -s {separators}       ["<< OptSeparators << "] Treat any of these characters as end of description.\n" <<
    "   -x                    ["<< (OptExclude? "true" : "false") << "] Exclude the listed reads?\n" <<
    "   -m                    ["<< (OptMultiSeq? "true" : "false") << "] Skip suppression of duplicate entries?\n" <<
    "   -n {SeqPerFile}       ["<< OptSeqPerFile << "] Put this many sequences in each output file, using\n"
    "                         -o option for prefix of multiple output files\n" <<
    "   -d {DebugString}      ["<< OptDebug << "] Each character turns on a debugging option;\n" <<
    "                            '+' turns on all debugging.\n" <<
    "\n";
}

//---------------------------------------------------
// * PrintHelp
//---------------------------------------------------
//
void PrintHelp()
{
  cout << 
    "atlas-extract-reads  Reads in files of read descriptions and uses that to extract\n" <<
    "                     reads out of a multi-FASTA file.  Works on FASTA sequence\n" <<
    "                     or quality -- run once on each file to do both.  Descriptions\n" <<
    "                     must be exact up to the first separator character (space by\n" <<
    "                     default).\n";
  PrintOptions("Default");
  cout << "Version: $Header: //CBT-depot/CompBioToolsApps/tools/atlas-extract-reads.cc#1 $\n";
}

//---------------------------------------------------
// * SetupOptions
//---------------------------------------------------
//
void SetupOptions(int argc, char**argv)
{
  // Create parser object and define valid options
  // Options with a : require a parameter. 
  // GetOption Parser(argc,argv,"i:f:o:d:n:s:mxh");

  // Default values
  OptFastaName = "";		// Fail if not on commmand line
  OptFileOfDescrip = "";        // Require this and/or SeqPerFile
  OptOutFile = "";		// Require this, or SeqPerFile, or put output on cout
  OptMultiSeq = false;
  OptExclude = false;
  OptSeqPerFile = 0;
  OptSeparators = " ";
  OptDebug = "";

  // Handle the options...
  int i;
  for (i = 1; i < argc; i++) {
    char prefix = argv[i][0];
    char theOption = argv[i][1];
    
    if (prefix == '-') {
      switch(theOption) {
      case 'i': OptFastaName = argv[++i]; break;
      case 'f': OptFileOfDescrip = argv[++i]; break;
      case 'o': OptOutFile = argv[++i]; break;
      case 'm': OptMultiSeq = true; break;
      case 'x': OptExclude = true; break;
      case 's': OptSeparators = argv[++i]; break;
      case 'n':
	OptSeqPerFile = atoi(argv[++i]);
	if (OptSeqPerFile < 1) {
	  cerr << "Sequences per file should be > 0.\n";
	  PrintHelp();
	  exit(1);
	}
	break;
      case 'd': OptDebug = argv[++i]; break;
      case 'h':
	PrintHelp(); myExit(0);
	break;
      default: 
	cerr << "Unrecognized option: " << theOption << endl;
	PrintHelp();
	goto EndOptions;
      }
    }
  }
    EndOptions:
  if (OptFastaName == ""){
    cerr << "No input FASTA file name specified -- stdin assumed.\n";
  }
  
  if (!OptSeqPerFile && (OptFileOfDescrip == "")) {
    cerr << "Must have a file of descriptions (-f) or sequences per file (-n) option.\n";
    PrintHelp();exit(1);
  } 
  if (OptOutFile == "") {
    if (OptFastaName != "") {
      if (OptSeqPerFile) {
	OptOutFile = OptFastaName + ".";
      } 
      else {
	OptOutFile = OptFastaName + "." + OptFileOfDescrip;
      }
    }
    else {
      if (OptSeqPerFile) {
	OptOutFile = "extract.";
      }
      else {
	OptOutFile = "extract.stdin";
      }
    }
  }
  
  if (debugging("o")) PrintOptions("Current");
}

// Hash stuff
struct eqstr
{
  bool operator()(const string& s1, const string& s2) const
  {
    return s1 == s2;
  }
};
struct hashstr {
  size_t operator()(const string s) const {
    hash<const char *> H;
    return H(s.c_str());
  }
};

class DescSet: public hash_set<string, hashstr, eqstr> {
};
DescSet descrips;

void hashDump(const char *msg) {
  DescSet::iterator dit;
  cout << "Hash table " << msg << endl;
  for (dit = descrips.begin(); dit != descrips.end(); ++dit) {
    cout << *dit << endl;
  }
}

void normalize(string &readName) {
  if (OptSeparators != "") {
    readName = readName.substr(0, readName.find_first_of(OptSeparators));
    if (debugging("r")) cerr << "Trimmed description: " << readName << endl;
    return;
  }
  for (int i = 0; i < readName.size(); i++) {
    if (isspace(readName[i])) {
      if ((i > 0) && (readName[i-1] == ' ')) {
	// delete second of repeating spaces
	readName.erase(i, i+1);
	// don't advance -- this compensates for later i++
	i--;
      }
      else { // make sure it's a plain space
	readName[i] = ' ';
      }
    }
  }
  // handle final space
  if (isspace(readName[readName.size()-1])) {
    readName.erase(readName.size()-1, readName.size());
  }
}

ifstream fod;

void readDescrips()
{
  string readName;
  while (getline(fod, readName)) {
    if (readName[0] == '>') {
      readName.erase(0, 1); // remove a leading '>'
    }
    normalize(readName);
    descrips.insert(readName);
    if (debugging("h")) {
      cerr << "Putting descrip " << readName << " into hash set." << endl;
      hashDump("");
    }
  }
}


void checkDescrip(const string &d)
{
  DescSet::const_iterator it = descrips.find(d);
  cout << d << "maps to: "
       << (it != descrips.end() ? *it : "not present")
       << endl;
}
bool lookupDescrip(const string &d)
{
  DescSet::const_iterator it = descrips.find(d);
  return (it != descrips.end());
}

// Output an entry to each of the bins where its readname was found.
// Output file name is constructed from OptBinPrefix, bin number, OptOutSuffix
// (note that OptOutSuffix has ".qual" appended before processing quality file)
//
ostream *output;

void printFasta(const string &faEntry) {
  static long seqCount = 0;
  string readName;
  istrstream entryStream(faEntry.c_str());
  getline(entryStream, readName);

  normalize(readName);
  if (OptFileOfDescrip != "") {
    if (debugging("r")) {
      cerr << readName << " " << descrips.count(readName) << endl;
      checkDescrip(readName);
    }
    if (!OptExclude && (descrips.count(readName) == 0)) {
      return;
    } 
    else if (OptExclude && (descrips.count(readName) > 0)) {
      return;
    }
  }

  // else 
  if (!seqCount || (OptSeqPerFile && ((seqCount % OptSeqPerFile) == 0))) {
    openNext();
  }
  seqCount++;
  (*output) << ">" << faEntry;

  if ((OptFileOfDescrip != "") && !OptMultiSeq && !OptExclude) {
    descrips.erase(readName);
  }
}

//------------------------------------------------------
// * main
//------------------------------------------------------
// 
int main(int argc,char**argv){

  SetupOptions(argc,argv);

  // Read descriptions
  if (OptFileOfDescrip != "") {
    fod.open(OptFileOfDescrip.c_str());
    if (debugging("r")) {
      cout << "\tReading descriptions file " << OptFileOfDescrip << endl;
    }
    readDescrips();
    fod.close();

    if (debugging("a")) hashDump(" after reading descrips:");
  }

  istream *input;
  ifstream inFile;
  string faEntry;
  faEntry.reserve(1<<19);	// reserve 512k for large reads without resize

  if (OptFastaName == "") {
    input = & cin;
  }
  else {

    inFile.open(OptFastaName.c_str());	// the multi-Fasta sequences
    input = & inFile;
  }

  cout << "Reading Fasta and copying to output..." << endl;
  if (!getline(*input, faEntry, '>') || faEntry.length()) {
    cerr << "Input format error for sequence file " << OptFastaName << endl;
    exit(1);
  }

  while (getline(*input, faEntry, '>')) {
    // Foreach FASTA entry (terminated by ">" or EOF)
    //	print it out if necessary
    printFasta(faEntry);
  }
  cout << "done." << endl;
  if ((OptFileOfDescrip != "") && debugging("a")) hashDump("after lookups:");

  if (OptFastaName != "") {
    inFile.close();
  }
  closeOut();
}

ofstream outFile;

static void openNext() {
  static int fileNum = 0;
  char fileNameBuf[100];
  if (OptSeqPerFile) {
    if (fileNum) {
      outFile.close();
    }
    sprintf(fileNameBuf, "%s%05d", OptOutFile.c_str(), fileNum);
    outFile.open(fileNameBuf);
  } 
  else if (!fileNum && (OptOutFile != "")) {
    outFile.open(OptOutFile.c_str());
  }
  fileNum++;
  if (OptOutFile != "") {
    output = &outFile;
  } else {
    output = &cout;
  }
}

static void closeOut() {
  if (OptOutFile != "") {
    outFile.close();
  }
}
