//==============================================================================
//	 
// atlas-screen-window.cc
//
//==============================================================================
// $Header: //CBT-depot/CompBioToolsApps/tools/atlas-screen-window.cc#1 $
//------------------------------------------------------------------------------
// $Log: atlas-screen-window.cc,v $
// Revision 1.4  2003/02/18 23:05:59  havlak
// Made sure combined trim length (both vector and quality trim) yields
// read length >= OptOkLen else screen the whole thing.
//
// Revision 1.3  2003/02/18 21:41:13  havlak
// Output sequence now uses Ns instead of Xs for masking.
// Option "-d C" prints out clipping points:
// readname	qstart	qend	vstart	vend
//
// Revision 1.2  2003/02/18 16:41:07  havlak
// Split off from atlas-window to do some more aggressive screening of vector and
// support Phrap better by being able to mask just on vector.
//
// Revision 1.1  2003/02/12 21:18:00  havlak
// Initial revision
//
// Revision 1.9  2001/11/30 17:10:54  havlak
// Removed the "INTERNAL ERROR" message.  The problem is rejection of
// otherwise passing windows because they are too short after removal of
// initial and terminal Xs and (when option +k is used) bases of low
// quality (usually < 20).  I want reads whose otherwise passing windows
// have these problems to be rejected -- since this is the desired
// behavior, the error message is unnecessarily confusing.
//
// Revision 1.8  2001/11/21 19:11:02  havlak
// Fixed major bug in writing out quality scores -- was repeating
// the same score for a whole line.
//
// Revision 1.7  2001/10/17 18:14:48  havlak
// More checking, options to control input screened, unscreened and
// quality file names.  Now trimming any leading or trailing Xs or
// zero-quality bases.
//
// Revision 1.6  2001/08/22 01:15:25  havlak
// Debugged and installed.  Added
//
// 	-v option to print RCS version information
// 	-c package of options for trimming contigs before crossmatch
//
// Revision 1.5  2001/08/21 16:42:47  havlak
// This version compiles again, more testing needed.
//
// Revision 1.4  2001/08/21 15:42:22  havlak
// Broken version in between adding options and changing name to
// atlas-window.
//
// Revision 1.3  2001/07/13 18:03:59  havlak
// Extensive changes to soft-screen contaminants -- requires both a screened copy and an
// unscreened copy of the sequence.
//
// Revision 1.2  2001/06/07 09:41:18  havlak
// Working version, reconciled with my version of Richard's MOU readlength
// perl script.
//
//------------------------------------------------------------------------------
//.......1.........2.........3.........4.........5.........6.........7.........8

#include "LocalSTL.hh"
#include <iomanip>
// #include "zipiostream.hh"
#include <math.h>
#include <ctype.h>

string OptDebug;
bool OptHavlak;
bool OptHardMask;
char OptScreenType;
bool OptPassedFile;
bool OptPassedQual;
bool OptOmitNone;
int OptOkQual;
int OptOkLen;
int OptWindow;
int OptPreVector;
double OptErrRate;
bool OptTrim;
string OptOrigSuffix;
string OptScreenSuffix;
string OptQualSuffix;
string OptVectorFile;

typedef vector<double> Errors;

class Clips {
public:
  int start;
  int end;
  Clips () {
    start = 0;
    end = -1;
  }
  Clips (int s, int e) {
    start = s;
    end = e;
  }
  int length() {
    return end + 1 - start;
  }
  void make_empty() {
    start = 0;
    end = -1;
  }
};

class ReadInfo {
public:
  string name;
  Clips clips;
  int rawHQ;
  int insert;
  int hq;
  string screened;
  string unscreen;
  vector<char> quals;
  string descrip;

  ReadInfo() {
    name = "";
    rawHQ = insert = hq = 0;
    unscreen = "";
    screened = "";
    descrip = "";
  }
  Clips clipQual(int okqual, int window, double okerr);
  Clips clipVec(int start, int end, int window);
  void print(ostream &, ostream&);
  void mask(Clips clipping);
  int length() { return clips.length(); }
};

void getStats(ReadInfo &w, int oklen, int okqual, int window, double okerr);
void getDescrip(string &f, string &d);
string getBases(string &f, string &b, bool asMasked);
string getQuals(string &f, vector<char> &q);
double sumError(Errors &e, int i, int window);

//**********************************************
//
// Definitions for the k-mer counting (10 to 16 bases)
//
typedef unsigned int Word;
const Word WORDLEN = 10;
const Word BASESUNUSED = (sizeof(Word) * 4 - WORDLEN);
const Word BITSPERBASE = 2;
const Word BITSUNUSED = (BASESUNUSED*BITSPERBASE);
const Word BITSPERWORD = (sizeof(Word) * 8);
const Word BASEMASK = ((1 << BITSPERBASE) - 1);
const Word BASESPERBYTE = (8/BITSPERBASE);
const Word BASESPERWORD = (BITSPERWORD/BITSPERBASE);
const Word A = 0;
const Word C = 1;
const Word G = 2;
const Word T = 3;
const Word NOGOOD = 0xFFFFFFFF;
const Word PAD = 0xFFFFFFFE;
const Word SHORT = 0xFFFFFFFD;
const int WORDTABLESIZE = 2 << (WORDLEN * 2);
unsigned char wordTable[WORDTABLESIZE] = { 0 };

static Word shiftin(char b) {
  static Word w;
  static Word r;
  static int ngood = 0;
  int code;

  switch (b) {
  case 'A': case 'a': {
    code = A;
    break;
  }
  case 'C': case 'c': {
    code = C;
    break;
  }
  case 'G': case 'g': {
    code = G;
    break;
  }
  case 'T': case 't': {
    code = T;
    break;
  }
  case ' ': case '\t': case '\n': case '*': {
    return PAD;
  }
  default: {
    ngood = 0;
    return NOGOOD;
  }
  }
  ngood++;
  w <<= BITSPERBASE + BITSUNUSED;
  w >>= BITSUNUSED;
  w |= code;
  r >>= BITSPERBASE;
  r |= (~code & BASEMASK) << (BITSPERWORD - (BITSPERBASE + BITSUNUSED));
  if (ngood < WORDLEN) return SHORT;
  return (w < r)? w : r;
}

void addVectorBase(char b) {
  Word result = shiftin(b);
  if (result < WORDTABLESIZE) wordTable[result] = 1;
}

bool checkVectorBase(char b) {
  Word result = shiftin(b);
  if (result < WORDTABLESIZE) return wordTable[result];
  else return 0;
}

void readVectorFile(const char *vf) {
  ifstream vecFile; // has vector for intensive screening in first OptPreVector bases
  string vecLine;

  vecFile.open(vf);

  while (getline(vecFile, vecLine)) {
    if (vecLine[0] == '>') {
      addVectorBase('>'); // just sets ngood to zero
    }
    else {
      for (int i = 0; i < vecLine.length(); i++) {
	addVectorBase(vecLine[i]);
      }
    }
  }
}

bool debugging(const char which[]) {
  if (OptDebug.find('+') != string::npos) return true;

  for (int i = 0; which[i]; i++) {
    if (OptDebug.find(which[i]) != string::npos) return true;
  }
  return false;
}

void ReadInfo::mask(Clips clipping) {
  int i;

  clips = clipping;

  for (i = 0; i < clips.start; i++) {
    screened[i] = OptHardMask? 'N' : tolower(unscreen[i]);
  }
  for (i = clips.start; i <= clips.end; i++) {
    screened[i] = toupper(unscreen[i]);
  }
  for (i = clips.end+1; i < screened.length(); i++) {
    screened[i] = OptHardMask? 'N' : tolower(unscreen[i]);
  }
}

void PrintOptions(string prefix) {
  cout << prefix << " option values are in []:\n"<<
    "   -O {string}      ["<< OptOrigSuffix <<"] suffix to append to arg for opening original (unscreened) sequence file\n"<<
    "   -S {string}      ["<< OptScreenSuffix <<"] suffix to append to arg for opening screened sequence file\n"<<
    "   -Q {string}      ["<< OptQualSuffix <<"] suffix to append to arg for opening quality file\n"<<
    "   -p {phredscore}  ["<< OptOkQual <<"] minimum phred quality score for good bases\n"<<
    "   -l {length}      ["<< OptOkLen <<"] minimum number of good bases in insert for passed read\n"<<
    "   -w {length}      ["<< OptWindow <<"] trim read from first passing window of this size to last\n"<<
    "   -P {length}      ["<< OptPreVector <<"] look extra hard for vector in this many bases at front of read\n" <<
    "   -e {error_rate}  ["<< OptErrRate <<"] maximum expected error rate for a passing window\n"<<
    "   -k or +k         ["<< (OptHavlak? "+":"-") <<"] trim to window bases with qual >= "<< OptOkQual <<"\n"<<
    "   -x or +x         ["<< (OptHardMask? "+":"-") <<"] + means trim/screen by hard masking (Xs) instead of soft (lowercase)\n"<<
    "   -s {char}        ["<< OptScreenType <<"] v = vector only, q = quality only, b = both, s = vector and short-qual reads\n" <<
    "   -f or +f         ["<< (OptPassedFile? "+":"-") <<"] create sequence file for processed reads\n"<<
    "   -q or +q         ["<< (OptPassedQual? "+":"-") <<"] create quality file for processed reads\n"<<
    "   -o or +o         ["<< (OptOmitNone? "+":"-") <<"] include all reads (not just passed) in output files\n"<<
    "   -t or +t         ["<< (OptTrim? "+":"-") <<"] trim by deleting bases outside windows instead of masking\n"<<
    "   -V {string}      ["<< OptVectorFile << "] path of file to read vector sequence from (for screening first 50 bases)\n" <<
    "   -P {length}      ["<< OptPreVector <<"] look extra hard for vector in this many bases at front of read\n" <<
    "   -d {DebugString} ["<< OptDebug <<"] Each character indicates a debugging option to turn on;\n"<<
    "                            '+' indicates turn on all debugging.\n"<<
    "                            'v' gives some statistics\n"<<
    "                            'V' gives lots of statistics\n"<<
    "   -v or +v         Print RCS version number\n"<<
    "   -h               print usage and option descriptions\n"
    ;
}


//---------------------------------------------------
// * PrintHelp
//---------------------------------------------------
//
void PrintHelp() {
  cout<<"$Id: //CBT-depot/CompBioToolsApps/tools/atlas-screen-window.cc#1 $\n"<<
    "atlas-screen-window  Reads in two fasta sequence files (masked and unmasked) and a\n" <<
    "       quality file for a set of reads to be trimmed.  Also requires a file of vector\n" <<
    "       sequences to be agressively trimmed (any 10bp match) at the beginnings of reads.\n" <<
    "       Looks for the first and last passing windows and masks any leading or trailing\n" <<
    "       junk.  Use with '-s v' to screen for Phrap without quality trimming.\n" <<
    "\n"<<
    "              Boolean options are specified with + to turn on, - to turn off (e.g., -x)\n"<<
    "\n";
  PrintOptions("Default");
  cout << "\nFor example,\n"<<
    "      atlas-screen-window +t -q -d 'Vo' foo.fa bar.fa\n";
}

//---------------------------------------------------
// * SetupOptions -- return index of first non-option argument
//---------------------------------------------------
//
int SetupOptions(int argc, char**argv)
{
  // Create parser object and define valid options
  // Options with a : require a parameter. 
  // GetOption Parser(argc,argv,"i:f:o:d:n:s:mxh");

  // Default values
  OptDebug = "v";
  OptScreenSuffix = "";
  OptOrigSuffix = "";
  OptQualSuffix = ".qual";
  OptHavlak = true;
  OptHardMask = false;
  OptScreenType = 'b';
  OptPassedFile = true;
  OptPassedQual = true;
  OptOmitNone = true;
  OptOkQual = 20;
  OptOkLen = 100;
  OptWindow = 50;
  OptPreVector = 50;
  OptErrRate = 1.25;
  OptTrim = false;
  OptVectorFile = "";

  // Handle the options...
  int i;
  for (i = 1; i < argc; i++) {
    char prefix = argv[i][0];
    char theOption = argv[i][1];
    if ((prefix == '-') || (prefix == '+')) {
      bool sense = (prefix == '+');
      switch(theOption) {
	case 'O': OptOrigSuffix = argv[++i]; break;
	case 'S': OptScreenSuffix = argv[++i]; break;
	case 'Q': OptQualSuffix = argv[++i]; break;
	case 'p': OptOkQual = atoi(argv[++i]); break;
	case 'l': OptOkLen = atoi(argv[++i]); break;
	case 'w': OptWindow = atoi(argv[++i]); break;
	case 'P': OptPreVector = atoi(argv[++i]); break;
	case 'e': OptErrRate = atof(argv[++i]); break;
	case 'k': OptHavlak = sense; break;
	case 'x': OptHardMask = sense; 
	  OptPassedFile = true;
	  break;
	case 's': OptScreenType = (argv[++i])[0];
	  break;
	case 'f': OptPassedFile = sense; break;
	case 'q': OptPassedQual = sense; break;
	case 'o': OptOmitNone = sense; break;
	case 't': OptTrim = sense; 
	  OptPassedFile = true;
	  break;
	case 'V': OptVectorFile = argv[++i]; break;
	case 'd': OptDebug = argv[++i]; break;
        case 'v': cout << "$Header: //CBT-depot/CompBioToolsApps/tools/atlas-screen-window.cc#1 $" << endl; break;
	default: 
	  cerr << "Unrecognized option: " << theOption << endl;
	case 'h': PrintHelp(); exit(0); break;
      }
    }
    else {
      // assume end of options and start of arguments
      break; // out of for loop
    }
  }
  if (debugging("o")) PrintOptions("Current ");
  return i;
}

//------------------------------------------------------
// * main
//------------------------------------------------------
// Analyze each of the argument FASTA files
//
int main(int argc,char**argv){

  // izipstream inFile;	// "command-line argument" contains base calls
  // izipstream qualFile;	// "command-line argument" + .qual
  ifstream inFile;	// "command-line argument" contains base calls
  ifstream qualFile;	// "command-line argument" + .qual
  ifstream origFile;    // "command-line argument" + .orig (unscreened sequence)

  ofstream outFile;	// <inFile>.passed         passed reads base calls
  ofstream oQualFile;	// <inFile>.passed.qual    passed reads quality scores

  string faEntry;	// grabs basecalls for one sequence
  string ofaEntry;	// grabs basecalls for one sequence
  string qEntry;	// grabs qualities for one sequence
  string fName, fRoot;  // hold names of input files
  bool readOrig;

  // Statistics over all files examined
  //
  // Read length is from start of first passing window to end of last
  // Insert length is number of non-screened bases in read length
  // HQ is number of bases of passing quality in the insert
  double readlength, readinsert, readHQ, readHQinsert;
  	// total readlength, insert, insert bases >= OptOkQual
  long readwincount, readcount, readHQcount;

  int firstArg = SetupOptions(argc, argv);
  readVectorFile(OptVectorFile.c_str());

  // total reads with passing windows, inserts, hq counts
  readwincount = readcount = readHQcount = 0;
  readlength = readinsert = readHQ = readHQinsert = 0.0;

  if (OptDebug == "v") {
    cout << "TraceRun_Name      passQual passPure Rbases Rq20 Ibases Iq20\n";
  }

  for (int i = firstArg; i < argc; i++) {
    fRoot = argv[i];

    fName = fRoot + OptScreenSuffix;
    if (debugging("f")) cout << "Screened file is " << fName << endl;
    inFile.open(fName.c_str());	// the multi-Fasta sequences

    fName = fRoot + OptQualSuffix;
    if (debugging("f")) cout << "Quality file is " << fName << endl;
    qualFile.open(fName.c_str());

    fName = fRoot + OptOrigSuffix;
    if (debugging("f")) cout << "Unscreened file is " << fName << endl;
    origFile.open(fName.c_str());

    // Eat the first > in the files.  Single | intended so both getlines
    // are called.
    //
    if ((!getline(inFile, faEntry, '>') | (!getline(qualFile, qEntry, '>')))
	|| faEntry.length() || qEntry.length()) {
      cerr << "Input format error for sequence file " << argv[i] << endl;
      cerr << "faEntry(" << faEntry.length() << "): " << faEntry << "/" << endl;
      cerr << "qEntry(" << qEntry.length() << "): " << qEntry << "/" << endl;
      exit(1);
    } 
    else {
      if (! getline(origFile, ofaEntry, '>') || ofaEntry.length()) {
	readOrig = false;
	cerr << "Unable to read original file for " << argv[i] << endl;
      }
      else {
	readOrig = true;
      }
      if (debugging("V")) {
	cout << "Reading sequence and quality for " << argv[i] << endl;
      }
    }

    if (OptPassedFile) {
      fName = fRoot + ".pass";
      outFile.open(fName.c_str());
      if (OptPassedQual) {
	fName += ".qual"; // <fRoot>.pass.qual
	oQualFile.open(fName.c_str());
      }
    }

    ReadInfo w;
    double filelength, fileinsert, fileHQ, fileHQinsert, fileHQlength, filerawHQ;
    long filewincount, filecount, fileHQcount;
    filewincount = filecount = fileHQcount = 0;	
    filerawHQ = filelength = fileinsert = fileHQ = fileHQinsert = fileHQlength = 0.0;

    // Grab entries up to the next > into faEntry, qEntry
    //
    bool skipping = false;
    while (getline(inFile, faEntry, '>')) {
      if (readOrig) {
	getline(origFile, ofaEntry, '>');
      }
      else {
	ofaEntry = faEntry;
      }
      if (skipping) {
	// skipping a sequence to see if that synchs us with quality
	skipping = false;
      }
      else {
	getline(qualFile, qEntry, '>');
      }

      getDescrip(faEntry, w.descrip);
      string sName = getBases(faEntry, w.screened, true);
      string uName = getBases(ofaEntry, w.unscreen, false);
      string qName = getQuals(qEntry, w.quals);

      if (uName != sName) {
	cerr << "Mismatch between screened and unscreened sequence names: " 
	     << sName << "/" << uName << endl;
	exit(1);
      }
      if (sName != qName) {
	cerr << "Mismatch between sequence and quality names: " 
	     << sName << "/" << qName << "; skipping to next sequence." << endl;
	skipping = true;
	continue;
      }
      w.name = sName;

      Clips qclip = w.clipQual(OptOkQual, OptWindow, OptErrRate);
      Clips vqclip = w.clipVec(qclip.start, qclip.end, OptWindow);
      Clips vclip = w.clipVec(0, w.unscreen.length() -1, OptWindow);
      if (vqclip.start > qclip.start) {
	vclip.start = vqclip.start;
      }
      if (vqclip.end < qclip.end) {
	vclip.end = vqclip.end;
      }
      if (debugging("C")) {
	cout << w.name << "\t" << qclip.start << "\t" << qclip.end << "\t" 
	     << vclip.start << "\t" << vclip.end << endl;
      }
      // Mask the non-passing regions
      Clips final;

      if (OptScreenType == 'q') { // quality trim
	final = qclip;
      }
      else if (OptScreenType == 'v') { // vector trim
	final = vclip;
      }
      else if (OptScreenType == 's') { // vector and x out quality-short reads
	final = vclip;
	if (qclip.length() < OptOkLen) {
	  final.make_empty();
	}
      }
      else { // both
	final.start = max(qclip.start, vclip.start);
	final.end = min(qclip.end, vclip.end);
	if (final.length() < OptOkLen) {
	  final.make_empty();
	}
      }

      w.mask(final);
      if (w.length() >= OptOkLen || OptOmitNone) {
	w.print(outFile, oQualFile);
      }
    }
    if (OptPassedFile) {
      outFile.close();
      if (OptPassedQual) {
	oQualFile.close();
      }
    }
  }
}

Clips ReadInfo::clipQual(int okQual, int window, double okerr) {
  Errors e;
  Clips r;
  int i, k;

  if (debugging("q")) {
    cout << "clipQual" << endl 
	 << "Descrip: " << descrip << endl
	 << "Quality: ";
    for (i = 0; i < quals.size(); i++) {
      cout << ((char) ('0' + quals[i]));
    }
    cout << endl;
  }
  e.clear();

  if (debugging("q")) cout << "Errors:   ";
  for (i = 0; i < quals.size(); i++) {
    if (unscreen[i] == 'N' || unscreen[i] == 'X') {
      e.push_back(1);
    }
    else {
      e.push_back(pow(10, -quals[i]/10.0));
    }
    if (debugging("r")) cout << i << '\t' << ((int) quals[i]) << '\t' << e[i] << endl;
    if (debugging("q")) {
      char c = (int) (e[i] * 50 + '0');
      cout << ((char) min(c, '~'));
    }
  }
  if (debugging("q")) cout << endl;

  double werror = okerr + 1; // make sure that if body doesn't execute we fail the read
  // Find first passing window
  int minqual = (OptHavlak? okQual : 1);
  for (k = 0; k <= screened.length() - window; k++) {
    if (quals[k] < minqual) {
      werror = okerr + 1;
      continue;
    }
    werror = sumError(e, k, window);
    // cout << k << "/" << werror << " ";
    if (werror <= okerr) break;
  }
  // cout << endl;

  if (werror > okerr) {
    r.start = 0;
    r.end = -1;
  }
  else {
    r.start = k;

    // Find last passing window
    for (k = screened.length()-1; k >= (window - 1); k--) {
      if (quals[k] < minqual)
	continue;
      werror = sumError(e, k - (window - 1), window);
      if (werror <= okerr) break;
    }
    r.end = k;

    if (r.length() < window) {
      // These "internal errors" occur when what would normally be a passing
      // window is too short because the bases just inside the boundaries have
      // qualities a bit below minqual (i.e., okqual when OptHavlak)
      // or are X'd out.
      // cerr << "INTERNAL ERROR: on " << stats.descrip 
      // << ": passed [" << stats.start << ".." << stats.end << "] length "
      // << stats.length << " < window length " << window << endl;
      r.start = 0;
      r.end = -1;
    }
  }
  return r;
}

Clips ReadInfo::clipVec(int s, int e, int window) {
  Clips r;
  int i, k;

  if (debugging("c")) {
    cout << "clipVec" << endl 
	 << "Descrip:  " << descrip << endl
	 << "Screened: " << screened << endl
	 << "Unscreen: " << unscreen << endl;
  }

  for (k = max(s + 1 - window, 0); k < OptPreVector; k++) {
    if (checkVectorBase(unscreen[k])) {
      screened[k] = 'X';
    }
  }

  int lastX = s-1; // set lastX in front of incoming range
  for (k = s; k <= e; k++) {
    if (screened[k] == 'X') {
      lastX = k;
      continue;
    }
    if (k - lastX >= window) break;
  }
  // cout << endl;

  if (k - lastX < window) {
    r.start = 0;
    r.end = -1;
  }
  else {
    r.start = lastX + 1;
    lastX = e+1; // set lastX after incoming range

    // Find last passing window
    for (k = e; k >= s; k--) {
      if (screened[k] == 'X') {
	lastX = k;
	continue;
      }
      if (lastX - k >= window) break;
    }
    r.end = lastX - 1;

    if (r.length() < window) {
      // These "internal errors" occur when what would normally be a passing
      // window is too short because the bases just inside the boundaries have
      // qualities a bit below minqual (i.e., okqual when OptHavlak)
      // or are X'd out.
      // cerr << "INTERNAL ERROR: on " << stats.descrip 
      // << ": passed [" << stats.start << ".." << stats.end << "] length "
      // << stats.length << " < window length " << window << endl;
      r.start = 0;
      r.end = -1;
    }
  }
  return r;
}

double sumError(Errors &e, int i, int window) {
  double errors = 0;
  for (int j = i; j < i+window; j++) {
    errors += e[j];
  }
  if (debugging("e")) cout << "Window error count: " << errors << endl;
  return errors;
}

void ReadInfo::print(ostream &oSeq, ostream &oQual) {
  int i, begin, after;

  if (OptTrim) {
    begin = clips.start;
    after = clips.end + 1;
  }
  else {
    begin = 0;
    after = screened.length();
  }

  if (begin >= after) return; // no sequence to print

  oSeq << '>' << descrip << endl;
  // 0..10
  // 5..7
  for (i = begin; i < after; i += 50) {
    oSeq << screened.substr(i, min(50,after-i)) << endl;
  }
  if (OptPassedQual) {
    oQual << '>' << descrip << endl;
    for (i = begin; i < after; i += 25) {
      for (int j = i; j < i + 25 && j < after; j++) {
	oQual << (int) quals[j] << " ";
      }
      oQual << endl;
    }
  }
}


void getDescrip(string &f, string &d) {
  istrstream s(f.c_str());
  getline(s, d, '\n');
}


string getBases(string &f, string &b, bool asMasked) {
  int i;
  string name = "";
  b = "";

  // skip to the first blank or newline
  for (i = 0; (i < f.length()) && (! isspace(f[i])); i++);
  name = f.substr(0, i);
  if (debugging("n")) cout << "Reading bases for " << name << "(asMasked=" 
			   << asMasked << ")" << endl;

  // skip to the newline...
  for (/* i = end of name */; (i < f.length()) && (f[i] != '\n'); i++);
  // and past
  i++;

  for (; i < f.length(); i++) {
    char letter;
    if (isalpha(f[i])) {
      letter = f[i];

      switch (letter) {
      case 'A': case 'T': case 'G': case 'C':
	break; // leave alone
      case 'a': case 't': case 'g': case 'c':
	if (asMasked) {
	  letter = 'N';
	}
	else {
	  letter = toupper(letter);
	}
	break;
      case 'X': case 'x': {
	if (asMasked) {
	  letter = 'X';
	}
	else {
	  letter = 'N';
	}
	break;
      }
      default:
	// other base possibilities are not counted
	letter = 'N';
	break;
      }
      b += letter;
    }
  }
  return name;
}


string getQuals(string &f, vector<char> &q) {
  string junk;
  int i, qval;
  string name;

  // skip to the first blank or newline
  for (i = 0; (i < f.length()) && (! isspace(f[i])); i++);
  name = f.substr(0, i);
  if (debugging("n")) cout << "Reading quals for " << name << endl;

  istrstream s(f.c_str());
  getline(s, junk, '\n');

  q.clear();
  i = 0;
  while (s >> qval) {
    q.push_back((char) qval);
  }
  return name;
}
