#include "CompBioTools.hh"
#include <string>
#include "zipiostream.hh"

// Field definitions for NCBI ancillary files. 
#define CENTER_NAME 3
#define INSERT_SIZE 16
#define INSERT_STDEV 17
#define SEQ_LIB_ID 30
#define TEMPLATE_ID 38
#define TI 39
#define TRACE_END 41


string OptFASTAFOF="";
string OptAncillaryFOF ="";
string OptRootPath = "";

string gVersion = "1.00";

//---------------------------------------------------
// * PrintVersion
//---------------------------------------------------
//
void PrintVersion()
{
  cout<<"
    Version "<<gVersion<<endl;
}

//---------------------------------------------------
// * PrintHelp
//---------------------------------------------------
//
void PrintHelp()
{
printf("\n
atlas-add-anc  -f FASTAFOF  -a AncillaryFOF  <-o RootPath> 

    Reads in all of the ancillary information from the ancillary files given 
    in AncillaryFOF.  Successively reads in the files specified in FASTAFOF, 
    rewriting the description line of each file to contain the corresponding 
    ancillary information.  The optional root path is a path that will 
    be added to any output files. 

    Written by: James Durbin (kdurbin@bcm.tmc.edu)
");

 PrintVersion();

 printf(" 
    Version History:  
    ----------------
    1.00
");
}

//---------------------------------------------------
// * SetupOptions
//---------------------------------------------------
//
void SetupOptions(int argc, char**argv)
{
  // Create parser object and define valid options
  // Options with a : require a parameter. 
  GetOption Parser(argc,argv,"f:a:o:h");

  // Handle the options...
  char theOption;
  while((theOption = Parser.NextOption()) != OPTION_EOF){
    switch(theOption){
	 case 'f': OptFASTAFOF = Parser.GetStringArg();break;
	 case 'a': OptAncillaryFOF = Parser.GetStringArg();break;
	 case 'o': OptRootPath = Parser.GetStringArg();break;
	 default: PrintHelp();exit(1);break;
    }
  }
}

//-------------------------------------------------------------------------------
// * AncillaryInfo
//-------------------------------------------------------------------------------
//
class AncillaryInfo{
public:
  string mCenterName;
  int    mInsert;
  int    mInsertStDev;
  string mTemplate;
  string mSeqLib;
  //int    mTraceID;
  char   mDirection;


  int  ParseLine(string line){
	 string temp;
	 CutString(line,mCenterName,CENTER_NAME,'\t');

	 CutString(line,temp,INSERT_SIZE,'\t');
	 mInsert = atoi(temp);

	 CutString(line,temp,INSERT_STDEV,'\t');
	 mInsertStDev = atoi(temp);
	 
	 CutString(line,mTemplate,TEMPLATE_ID,'\t');

	 CutString(line,mSeqLib,SEQ_LIB_ID,'\t');

	 CutString(line,temp,TRACE_END,'\t');
	 mDirection = temp[0];

	 // Add a few defaults where missing. -----


	 // KJD Note:  Some defaults are added at in AddAncillaryToDescription

	 if (mSeqLib==""){
		if (mCenterName == "BCM"){
		  mSeqLib = mTemplate.substr(0,5);
		}else if (mCenterName == "GTC"){
		  mSeqLib = "GTC_DUMMY";
		}
	 }

	 if (mInsert <= 0){
		if (mCenterName == "BCM"){
		  mInsert = 3000;
		  mInsertStDev = 2000;
		}else if(mCenterName == "GTC"){
		  mInsert = 2000;
		  mInsertStDev = 1000;
		}else{
		  mInsert=3000;
		  mInsertStDev = 2000;
		}
	 }

	 if (mInsertStDev <= 0){
		int  TenPercent = (int)(0.10 * (float)mInsert);
		int  maxVal = max(TenPercent,500);
		mInsertStDev = maxVal;
	 }
	 // --------------

	 // Get the traceID, to be used as a key in the hash, but otherwise
	 // not needed. 
	 CutString(line,temp,TI,'\t');
	 int TraceID = atoi(temp);
	 return(TraceID);

  }
};

//-------------------------------------------------------------------------------
// * AncillaryInfoVec
//-------------------------------------------------------------------------------
// A vector of ancillary information wth a map between the trace index and the 
// corresponding ancillary information. 
//
class AncillaryInfoVec : public vector<AncillaryInfo>{
public:
  Int2IntMap mTIMap;
  
  //-------------------------------------------
  // * Append
  //-------------------------------------------
  //
  void Append(istream &fin){
	 string line;
	 int lineCount = 0;
	 do{
		fin.getline(readbuf,READ_BUFFER_SIZE);
		line = readbuf;
		lineCount++;
		if (line != ""){
		  int newIdx = size();
		  resize(newIdx+1);
		  int TraceID = (*this)[newIdx].ParseLine(line);
		  mTIMap[TraceID] = newIdx;
		}
	 }while(!fin.eof());
	 cout<<" lineCount= "<<lineCount<<" ";
  }


  //-------------------------------------------
  // * Read
  //-------------------------------------------
  //
  void Read(string &FName){
	 izipstream ancfile;
	 ancfile.open(FName.c_str());
	 if (ancfile.fail()){
		cout<<"Failed to open "<<FName<<endl;
		return;
	 }

	 Append(ancfile);

	 ancfile.close();
  }


  //-------------------------------------------
  // * ReadMultipleFiles
  //-------------------------------------------
  //
  void ReadMultipleFiles(string &FOFName){

	 izipstream alist;
	 alist.open(FOFName.c_str());
	 if (alist.fail()){cout<<"Failed to open "<<FOFName<<endl;exit(1);}

	 do{
		string FileName;
		GetLine(alist,FileName);		
		
		if (FileName != ""){
		  izipstream ancfile;
		  ancfile.open(FileName.c_str());
		  if (ancfile.fail()){cout<<"Failed to open "<<FOFName<<endl;continue;}
		
		  cout<<"\t Adding "<<FileName<<".."<<flush;
		  Append(ancfile);
		  cout<<"done. size="<<size()<<endl;
	  
		  ancfile.close();
		}
	 }while(!alist.eof());
  }

};

//-------------------------------------------
// * AddAncillaryToDescription
//-------------------------------------------
//
void AddAncillaryToDescription(AncillaryInfo &A,string &Description)
{
  // The old description contains the trace name.  Extract that as 
  // the beginning of the new trace name.  
  int offset = 0;
  Description = GetNextWSDelimitedString(Description,offset,offset);
  string CenterReadName = Description;

  // Add some default template information for Baylor reads when missing
  if ((A.mTemplate == "") || (ContainsOnlyWS(A.mTemplate))){
	 if (A.mCenterName=="BCM"){
		// Name in format XXXXX9C12345 -> XXXXX12345
		// Sometimes a letter is mistakenly on the end, thus I use GetNextInteger
		// to get the integer part out ignoring any subsequent letters. 
		string Root = CenterReadName.substr(0,5);
		int offset = 6;
		int Tail = GetNextInteger(CenterReadName,offset,offset);		
		string TailStr=itoa(Tail);
		A.mTemplate = Root+TailStr;
	 }
  }

  // Rewrite description line with this info. 
  //	 SeqVec[i].mDescription += " CENTER_NAME: "+A.mCenterName;
  // Presumes the description contains the read name, so just add additional info. 

  if ((A.mCenterName != "") && (!ContainsOnlyWS(A.mCenterName))){
	 Description += " CENTER: ";
	 Description += A.mCenterName;
  }

  if ((A.mTemplate != "") && (!ContainsOnlyWS(A.mTemplate))){
	 Description += " TEMPLATE: "+A.mTemplate;
  }  

  if (toupper(A.mDirection) == 'F'){
	 Description += " DIRECTION: fwd";
  }else if (toupper(A.mDirection) == 'R'){
	 Description += " DIRECTION: rev";
  }
  
  if (A.mInsert != 0){
	 Description += " INSERT_SIZE: "+itoa(A.mInsert);
	 Description += " INSERT_STDEV: "+itoa(A.mInsertStDev);
  }

  if ((A.mSeqLib!="") && (!ContainsOnlyWS(A.mSeqLib))){
	 Description += " LIBRARY: "+A.mSeqLib;
  }
  
  //SeqVec[i].mDescription = "TI: "+itoa(A.mTraceID);
  
  //SeqVec[i].mDescription += " OLD_DESC: ";
  //SeqVec[i].mDescription += oldDesc;
}

//-------------------------------------------
// * AddAncillaryToDescriptions
//-------------------------------------------
//
void AddAncillaryToDescriptions(AncillaryInfoVec &AVec,SequenceVector &SeqVec)
{
  for(int i = 0;i< SeqVec.size();i++){
	 //string oldDesc = SeqVec[i].mDescription;

	 // Find the right ancillary info
	 string TIStr = SeqVec[i].mName.substr(7,SeqVec[i].mName.size()-7);
	 int TIVal = atoi(TIStr);
	 int AncIdx = AVec.mTIMap[TIVal];
	 AncillaryInfo &A = AVec[AncIdx];

	 AddAncillaryToDescription(A,SeqVec[i].mDescription);

  }
}

//-------------------------------------------
// * AddAncillaryToDescriptions
//-------------------------------------------
//
void AddAncillaryToDescriptions(AncillaryInfoVec &AVec,SequenceQualityVector &SeqQualVec)
{
  for(int i = 0;i< SeqQualVec.size();i++){
	 //string oldDesc = SeqVec[i].mDescription;

	 // Find the right ancillary info
	 string TIStr = SeqQualVec[i].mName.substr(7,SeqQualVec[i].mName.size()-7);
	 int TIVal = atoi(TIStr);
	 int AncIdx = AVec.mTIMap[TIVal];
	 AncillaryInfo &A = AVec[AncIdx];

	 AddAncillaryToDescription(A,SeqQualVec[i].mDescription);
  }
}



//-------------------------------------------
// * main
//-------------------------------------------
//
int main(int argc,char **argv)
{
  FASTAFile FA;

  SetupOptions(argc,argv);

  // Read in all of the ancillary information. 
  long StartTime;
  StartTiming(StartTime);
  cout<<"Reading ancillary information..."<<endl;
  AncillaryInfoVec AInfo;
  AInfo.ReadMultipleFiles(OptAncillaryFOF);
  float elapsed = EndTiming(StartTime);
  cout<<"Elapsed time = "<<elapsed<<endl;

  // Get the list of FASTA files to process.
  vector<string> FASTANames;
  ReadFOF(OptFASTAFOF,FASTANames);

  cout<<"Processing FASTA files..."<<endl;
  for(int fIdx = 0;fIdx<FASTANames.size();fIdx++){
	 SequenceVector *pSeqVec;
	 pSeqVec = new SequenceVector();

	 SequenceVector & SeqVec = *pSeqVec;

	 string FAName = FASTANames[fIdx];
	 cout<<"\t"<<FAName<<endl;

	 int dotIdx = FAName.find('.');
	 string FARoot = FAName.substr(dotIdx+1,FAName.size()-dotIdx-1);

	 // Chop off suffix if it's gz since the output will be plaintext.
	 int lastDotIdx = FARoot.rfind('.');
	 string suffix = FARoot.substr(lastDotIdx+1,FARoot.size()-lastDotIdx-1);
	 if (suffix =="gz"){
		FARoot = FARoot.substr(0,lastDotIdx);
	 }


	 //cout<<"FARoot="<<FARoot<<endl;

	 // Read FASTA input, add ancillary information. 
	 izipstream zin;
	 zin.open(FAName.c_str());
	 cout<<"\t\tReading..."<<flush;
	 FA.Read(zin,SeqVec);
	 cout<<"done."<<endl;
	 zin.close();
	 cout<<"\t\tAdding ancillary..."<<flush;
	 AddAncillaryToDescriptions(AInfo,SeqVec);
	 cout<<"done."<<endl;

	 //	 ozipstream zout;
	 ofstream zout;
	 string FAOutName = OptRootPath + "fasta.";
	 FAOutName+=FARoot;
	 cout<<"\t\tWriting to "<<FAOutName<<"..."<<flush;
	 zout.open(FAOutName.c_str());
	 if (zout.fail()){cout<<FAOutName<<" open failed"<<endl;exit(1);}
	 FA.Write(zout,SeqVec);
	 zout.close();
	 cout<<"done."<<endl;

	 delete pSeqVec;

	 /*

	 // These are done separately to conserve space. 
	 SequenceQualityVector SeqQualVec;
	 string qualInName = "qual.";
	 qualInName+=FARoot;
	 qualInName+=".gz";
	 cout<<"\t"<<qualInName<<endl;

	 cout<<"\t\tReading..."<<flush;
	 zin.open(qualInName.c_str());
	 FA.ReadQual(zin,SeqQualVec);
	 zin.close();
	 cout<<"done. Size="<<SeqQualVec.size()<<endl;

	 cout<<"\t\tAdd ancillary..."<<flush;
	 AddAncillaryToDescriptions(AInfo,SeqQualVec);
	 cout<<"done."<<endl;

	 string qualOutName = OptRootPath+"qual.";
	 qualOutName+=FARoot;
	 cout<<"\t\t Writing "<<qualOutName<<"..."<<flush;
	 zout.open(qualOutName.c_str());
	 if (zout.fail()){cout<<qualOututName<<" open failed"<<endl;exit(1);}
	 FA.WriteQual(zout,SeqQualVec);
	 zout.close();
	 cout<<"done."<<endl;
	 */

	 

  }
}











